#ifndef FLOWJO_EXCEPTION_HPP
#define FLOWJO_EXCEPTION_HPP

#include <exception>
#include <string>

namespace FJ
{
   /*
   * This is the base class for FlowJo exceptions. General FlowJo exceptions,
   * such as not being able to open a file, not being able to allocate memory, etc.
   * should derive from this class. Exceptions specific to a project should
   * also derive from this class but be put into a namespace for that project
   * 
   * Created by Greg Reese on February 15, 2022
   */
   class FlowJo_Exception : public std::exception
   {
   public:
      FlowJo_Exception( const char* text = nullptr );
      FlowJo_Exception( const std::string& text );

      // standard exceptions derived from std::exception, such as std::run_time and std::logic_error do not 
      // allow their copy constructors and assignment operators to throw exceptions from C++11 on, 
      // i.e. they're declared noexcept
      // We allow that here because it would be rare for them to actually throw an exception
      // and it makes it much easier to implement
      FlowJo_Exception( const FlowJo_Exception& source ) = default;
      FlowJo_Exception& operator=( const FlowJo_Exception& rhs ) = default;

      void append( const std::string& text, bool add_space = true );
      // appends the passed text to the stored text. Adds a space
      // before the appended text if add_space is true

      void what( const std::string& new_what );
      // allow the user to set text after the exception has been constructed
      // or during the construction of a derived class

      virtual const char* what() const noexcept;
      // more info on error that occurred

   private:

      std::string m_text;
   };

} // end namespace FJ
#endif