//
//  JsonIncludes.h
//  Native Engine
//
//  Created by Clayton Simons on 3/15/18.
//

#ifndef JsonIncludes_h
#define JsonIncludes_h

// on windows wingdi.h contains a macro that defines GetObject
#ifdef WINDOWS
#ifdef GetObject
#undef GetObject
#endif
#endif

#include <string>
#include <iostream>
#define RAPIDJSON_HAS_STDSTRING 1
//#define RAPIDJSON_HAS_CXX11_NOEXCEPT 0
#define RAPIDJSON_ASSERT(CONDITION) { if(!(CONDITION)) { throw EngineException(ExceptionType::Identifier::ParseError, "RAPIDJSON_ASSERT failed: " #CONDITION, explain_where(""), Error_PARSE_JSON); } }
#define RAPIDJSON_ASSERT_THROWS 1
#include "rapidjson/document.h"         // rapidjson's DOM-style API
#include "rapidjson/prettywriter.h"     // for stringify JSON
#include "rapidjson/pointer.h"
#include "rapidjson/allocators.h"
#include "rapidjson/schema.h"
#include "rapidjson/error/en.h"

#include "Exceptions.h"

using namespace rapidjson;
using rapidjson::internal::Hasher;
using std::string;

typedef MemoryPoolAllocator<CrtAllocator> AllocatorType;
typedef GenericValue<UTF8<>,MemoryPoolAllocator<CrtAllocator>> JSONValue;
typedef void (*ffiCallbackFunction)(char*);

//*********************************Start of Utility Methods*****************************
static inline string Stringify(const Value& v)
{
    StringBuffer buffer;
#ifdef FJ_DEBUG
    PrettyWriter<StringBuffer> writer(buffer);
#else
	Writer<StringBuffer> writer(buffer);
#endif
    v.Accept(writer);
    return buffer.GetString();
}

//commutative for arrays meaning order doesn't matter
static inline uint64_t GetCommutativeHash(const Value& v)
{
    MemoryPoolAllocator<> a;
	 Hasher<Value::EncodingType, MemoryPoolAllocator<>> hasher( &a, 256, false );
	 v.Accept(hasher);
    uint64_t answer = hasher.GetHashCode();
    return answer;
}

//equivalent if order doesn't matter for arrays
static inline bool IsEquivalent(const Value& v1, const Value& v2)
{
    return GetCommutativeHash(v1) == GetCommutativeHash(v2);
}

//Normal hash where order matters for arrays
static inline uint64_t GetHash(const Value& v)
{
	MemoryPoolAllocator<> a;
	 Hasher<Value::EncodingType, MemoryPoolAllocator<>> hasher( &a, 256, true );
	 v.Accept(hasher);
	uint64_t answer = hasher.GetHashCode();
	return answer;
}

//Test that two jsons are the same, including order of arrays
static inline bool IsSameJSON(const Value& v1, const Value& v2)
{
	return GetHash(v1) == GetHash(v2);
}

namespace std
{
	template<> struct hash<JSONValue>
	{
		size_t operator()(JSONValue const& s) const noexcept
		{
			return (size_t) GetCommutativeHash(s);
		}
	};
}

static inline JSONValue* GetValueByPointerChecked(JSONValue& root, const string& pointer)
{
	Pointer p(pointer);
	size_t unresolvedTokenIndex;
    JSONValue* member = GetValueByPointer(root, p, &unresolvedTokenIndex);
	if(member == nullptr)
	{
		stringstream msg;
		const Pointer::Token* tokens = p.GetTokens();
		string resolved = "";
		for(size_t i = 0; i < unresolvedTokenIndex; i++)
		{
			resolved += string("/") + tokens[i].name;
		}
		msg << "\"" << tokens[unresolvedTokenIndex].name << "\" was not found in \"" << resolved << "\".";
		throw EngineException(ExceptionType::Identifier::out_of_range, msg.str().c_str());
	}
    return member;
}

static inline const JSONValue* GetValueByPointerChecked(const JSONValue& root, const string& pointer)
{
	Pointer p(pointer);
	size_t unresolvedTokenIndex;
	const JSONValue* member = GetValueByPointer(root, p, &unresolvedTokenIndex);
	if(member == nullptr)
	{
		stringstream msg;
		const Pointer::Token* tokens = p.GetTokens();
		string resolved = "";
		for(size_t i = 0; i < unresolvedTokenIndex; i++)
		{
			resolved += string("/") + tokens[i].name;
		}
		msg << "\"" << tokens[unresolvedTokenIndex].name << "\" was not found in \"" << resolved << "\".";
		throw EngineException(ExceptionType::Identifier::out_of_range, msg.str().c_str());
	}
    return member;
}

static inline string GetStringByPointer(const JSONValue& root, const string& pointer)
{
	const JSONValue* member = GetValueByPointerChecked(root, pointer);
	if(!member->IsString())
		throw EngineException(ExceptionType::Identifier::invalid_argument, (pointer + " is not a string.").c_str());
    return member->GetString();
}

static inline int GetIntegerByPointer(const JSONValue& root, const string& pointer)
{
	const JSONValue* member = GetValueByPointerChecked(root, pointer);
	if(!member->IsInt())
		throw EngineException(ExceptionType::Identifier::invalid_argument, (pointer + " is not an integer.").c_str());
    return member->GetInt();
}

static inline Value GetObjectByPointer(JSONValue& root, const string& pointer)
{
	JSONValue* member = GetValueByPointerChecked(root, pointer);
	if(!member->IsObject())
		throw EngineException(ExceptionType::Identifier::invalid_argument, (pointer + " is not an object.").c_str());
	Value object(kObjectType);
	object = member->GetObject();
	return object;
}

static inline JSONValue::ConstObject GetObjectByPointer(const JSONValue& root, const string& pointer)
{
	const JSONValue* member = GetValueByPointerChecked(root, pointer);
	if(!member->IsObject())
		throw EngineException(ExceptionType::Identifier::invalid_argument, (pointer + " is not an object.").c_str());
	return member->GetObject();
}

static inline Value GetArrayByPointer(JSONValue& root, const string& pointer)
{
	JSONValue* member = GetValueByPointerChecked(root, pointer);
	if(!member->IsArray())
		throw EngineException(ExceptionType::Identifier::invalid_argument, (pointer + " is not an array.").c_str());
	Value array(kArrayType);
	array = member->GetArray();
	return array;
}

static inline JSONValue::ConstArray GetArrayByPointer(const JSONValue& root, const string& pointer)
{
	const JSONValue* member = GetValueByPointerChecked(root, pointer);
	if(!member->IsArray())
		throw EngineException(ExceptionType::Identifier::invalid_argument, (pointer + " is not an array.").c_str());
	return member->GetArray();
}

static inline bool GetBoolByPointer(const JSONValue& root, const string& pointer)
{
	const JSONValue* member = GetValueByPointerChecked(root, pointer);
	if(!member->IsBool())
		throw EngineException(ExceptionType::Identifier::invalid_argument, (pointer + " is not a boolean.").c_str());
    return member->GetBool();
}

static inline float GetFloatByPointer(const JSONValue& root, const string& pointer)
{
	const JSONValue* member = GetValueByPointerChecked(root, pointer);
	if(!member->IsNumber())
		throw EngineException(ExceptionType::Identifier::invalid_argument, (pointer + " is not a number.").c_str());
	return member->GetFloat();
}

static inline double GetDoubleByPointer(const JSONValue& root, const string& pointer)
{
	const JSONValue* member = GetValueByPointerChecked(root, pointer);
	if(!member->IsNumber())
		throw EngineException(ExceptionType::Identifier::invalid_argument, (pointer + " is not a number.").c_str());
	return member->GetDouble();
}

template <typename T>//Overloaded function for SetValueByPointer
typename T::ValueType& SetValueByPointer(T& root, const std::basic_string<typename T::Ch>& source, const std::basic_string<typename T::Ch>& value, typename T::AllocatorType& a) {
    return GenericPointer<typename T::ValueType>(source).Set(root, value, a);
}
template <typename DocumentType>//Overloaded function for SetValueByPointer
typename DocumentType::ValueType& SetValueByPointer(DocumentType& document, const std::basic_string<typename DocumentType::Ch>& source, const std::basic_string<typename DocumentType::Ch>& value) {
    return GenericPointer<typename DocumentType::ValueType>(source).Set(document, value);
}

// These methods extract the value for a given name, with sane defaults (and no assertion) if it's missing
static inline const char* StringValue(const JSONValue& object, const char* name, const char* defaultValue="") {
	auto it = object.FindMember(name);
	if(it == object.MemberEnd() || !it->value.IsString())
		return defaultValue;
	return it->value.GetString();
}
static inline bool BoolValue(const JSONValue& object, const char* name, bool defaultValue=false) {
	auto it = object.FindMember(name);
	if(it == object.MemberEnd() || !it->value.IsBool())
		return defaultValue;
	return it->value.GetBool();
}
static inline int IntValue(const JSONValue& object, const char* name, int defaultValue=0) {
	auto it = object.FindMember(name);
	if(it == object.MemberEnd() || !it->value.IsInt())
		return defaultValue;
	return it->value.GetInt();
}
static inline float FloatValue(const JSONValue& object, const char* name, float defaultValue=0) {
	auto it = object.FindMember(name);
	if(it == object.MemberEnd() || !it->value.IsFloat())
		return defaultValue;
	return it->value.GetFloat();
}
static inline double DoubleValue(const JSONValue& object, const char* name, double defaultValue=0) {
	auto it = object.FindMember(name);
	if(it == object.MemberEnd())
		return defaultValue;
    if (it->value.IsInt())
        return (double)it->value.GetInt();
    if (it->value.IsFloat())
        return (double)it->value.GetFloat();
    if (it->value.IsDouble())
        return it->value.GetDouble();
    return defaultValue;
}

// These methods offer a uniform interface for creating JSONValues, since the GenericValue constructors do not. This allows them to be used below in a templated way where the type of val is not known.
template<typename value_type>
static inline JSONValue AsJSONValue(const value_type& val, AllocatorType& a)
{
	return JSONValue(val);
}

template<>
inline JSONValue AsJSONValue<string>(const string& val, AllocatorType& a)
{
	return JSONValue(val, a);
}

//template<>
//inline JSONValue AsJSONValue<const string>(const string& val, AllocatorType& a)
//{
//	return JSONValue(val, a);
//}

template<>
inline JSONValue AsJSONValue<JSONValue>(const JSONValue& val, AllocatorType& a)
{
	return JSONValue(val, a);
}

//template<>
//inline JSONValue AsJSONValue<const JSONValue>(const JSONValue& val, AllocatorType& a)
//{
//	return JSONValue(val, a);
//}

// This method allows generic conversion from a container of values to a JSON array, with optional duplicate removal.
template<typename Iterator>
static inline JSONValue ToJSONArray(Iterator it, Iterator end, AllocatorType& a, bool removeDuplicates)
{
	unordered_set<JSONValue> seen;

	JSONValue ret(kArrayType);
	for(; it != end; ++it)
	{
		if(removeDuplicates)
		{
			if(!seen.insert(AsJSONValue(*it, a)).second)
				continue;
		}
		ret.PushBack(AsJSONValue(*it, a), a);
	}
	return ret;
}

// This method allows generic conversion from a container of values to a JSON array, with a function passed in to extract each value, and with optional duplicate removal.
template<typename Iterator, typename Fun>
static inline JSONValue ToJSONArray(Iterator it, Iterator end, const Fun& f, AllocatorType& a, bool removeDuplicates)
{
	unordered_set<JSONValue> seen;
	
	JSONValue ret(kArrayType);
	for(; it != end; ++it)
	{
		auto val = f(*it);
		if(removeDuplicates)
		{
			if(!seen.emplace(AsJSONValue(val, a)).second)
				continue;
		}
		ret.PushBack(AsJSONValue(val, a), a);
	}
	return ret;
}

static inline vector<string> ToStringVector(const JSONValue::ConstArray& jsonStringArray)
{
	vector<string> ret;
	for(const JSONValue& entry : jsonStringArray)
		ret.push_back(entry.GetString());
	return ret;
}

static inline bool ArrayHasString(const JSONValue& array, const string value)
{
	assert(array.IsArray());
	for(rapidjson::Value::ConstValueIterator itr = array.Begin(); itr != array.End(); ++itr)
	{
		string itrStr = itr->GetString();
		if( itrStr.compare( value) == 0)
			return true;
	}
	return false;
}
static inline bool ArrayHasDouble(const JSONValue& array, const double value)
{
	assert(array.IsArray());
	for(rapidjson::Value::ConstValueIterator itr = array.Begin(); itr != array.End(); ++itr)
	{
		if(itr->GetDouble() == value)
			return true;
	}
	return false;
}
#endif /* JsonIncludes_h */
