#include "Line_Counter.hpp"

#include "flowjo_utilities.hpp"

#include <cctype>
#include <stdexcept>
#include <fstream>

using namespace FJ;
using namespace std;

Line_Counter::Line_Counter( std::string file_name )
{
   if( !file_name.empty() )
      count( file_name );
}

int Line_Counter::num_blank_lines() const
{
   return m_file_counts.num_blanks;
}

int Line_Counter::num_classes() const
{
   return m_file_counts.num_classes;
}

int Line_Counter::num_code_lines() const
{
   return m_file_counts.num_code;
}

int Line_Counter::num_comment_lines() const
{
   return m_file_counts.num_long_comments
      + m_file_counts.num_short_comments;
}

int Line_Counter::num_lines() const
{
   return num_noncode_lines() + num_code_lines();
}

int Line_Counter::num_long_comment_lines() const
{
   return m_file_counts.num_long_comments;
}

int Line_Counter::num_noncode_lines() const
{
   return num_blank_lines() + num_comment_lines();
}

int Line_Counter::num_short_comment_lines() const
{
   return m_file_counts.num_short_comments;
}

int Line_Counter::num_structs() const
{
   return m_file_counts.num_structs;
}

float Line_Counter::ratio_code_noncode() const
{
   return num_noncode_lines() != 0
      ? static_cast<float>(num_code_lines()) / num_noncode_lines() : m_invalid;
}

int Line_Counter::total_blank_lines() const
{
   return m_total_counts.num_blanks;
}

int Line_Counter::total_classes() const
{
   return m_total_counts.num_classes;
}

int Line_Counter::total_code_lines() const
{
   return m_total_counts.num_code;
}

int Line_Counter::total_comment_lines() const
{
   return m_total_counts.num_long_comments 
      + m_total_counts.num_short_comments;
}

int Line_Counter::total_files() const
{
   return m_total_counts.num_files;
}

int Line_Counter::total_lines() const
{
   return total_noncode_lines() + total_code_lines();
}
int Line_Counter::total_long_comment_lines() const
{
   return m_total_counts.num_long_comments;
}

int Line_Counter::total_noncode_lines() const
{
   return total_blank_lines() + total_comment_lines();
}

float Line_Counter::total_ratio_code_noncode() const
{
   return total_noncode_lines() != 0 
      ? static_cast<float>(total_code_lines()) / total_noncode_lines() : m_invalid;
}

int Line_Counter::total_short_comment_lines() const
{
   return m_total_counts.num_short_comments;
}

int Line_Counter::total_structs() const
{
   return m_total_counts.num_structs;
}

void Line_Counter::clear_totals()
{
   m_total_counts.num_blanks = 0;
   m_total_counts.num_code = 0;
   m_total_counts.num_files = 0;
   m_total_counts.num_long_comments = 0;
   m_total_counts.num_short_comments = 0;
}

bool Line_Counter::is_blank_line( const std::string& line ) const
{
   return line.empty();
}
// return true iff line is empty

bool Line_Counter::is_class_line( const std::string& line ) const
{
   const static auto class_len { string( "class " ).size() };
   return line.size()>=class_len 
      && std::isspace( static_cast<unsigned char>(line[class_len-1]) ) 
      && line.substr(0,class_len-1)=="class"
      && line.find( ';' ) == string::npos;
}
// return true iff the first word is "class" and there is no
// semicolon in the line. (This prevents a forward declaration
// from being counted as a class line).

bool Line_Counter::is_comment_end_line( const std::string& line ) const
{
   return line.size() >= 2 && line.substr( line.size()-2, 2 ) == "*/";
}

bool Line_Counter::is_comment_line( const std::string& line ) const
{
   return line.size() >= 2 && line.substr( 0, 2 ) == "//";
}

bool Line_Counter::is_comment_start_line( const std::string& line ) const
{
   return line.size() >= 2 && line.substr( 0, 2 ) == "/*";
}

bool Line_Counter::is_struct_line( const std::string& line ) const
{
   const static auto struct_len { string( "struct " ).size() };
   const bool a = line.size() >= struct_len;
   if( !a )
      return a;
   const bool b = std::isspace( static_cast<unsigned char>(line[struct_len - 1]) );
   if( !b )
      return b;
   bool c = line.substr( 0, struct_len - 1 ) == "struct";
   if( !c )
      return c;
   bool d = line.find( ';' ) == string::npos;
   return a && b && c && d;
}
// return true iff the first word is "struct" and there is no
// semicolon in the line. (This prevents a forward declaration
// from being counted as a class line).

void Line_Counter::count( const std::string& file_name )
{
   vector<string> lines { make_lines( file_name ) };
   count( lines );
}

void Line_Counter::count( const std::vector<std::string> & lines )
{
   m_file_counts.clear();
   string line;
   bool in_comment { false };
   for( auto line : lines )
   {
      line = trim( line );

      // all is_*_line() require a trimmed line for input
      if( in_comment || is_comment_start_line( line ) )
      {
         in_comment = true; // may not be set if start line
         ++m_file_counts.num_long_comments;
         if( is_comment_end_line(line) )
            in_comment = false;
      }

      // not in a long comment for rest of these
      else if( is_blank_line( line ) )
         ++m_file_counts.num_blanks;

      else if( is_comment_line( line ) )
         ++m_file_counts.num_short_comments;

      else if( is_class_line( line ) )
         ++m_file_counts.num_classes;

      else if( is_struct_line( line ) )
         ++m_file_counts.num_structs;

      else // it's gotta be code
         ++m_file_counts.num_code;
   } // end of for loop

   // didn't count structs and classes as lines of code
   // so do that now
   m_file_counts.num_code +=
      m_file_counts.num_classes + m_file_counts.num_structs;

   // now add stuff for this file into totals
   m_total_counts += m_file_counts;

   ++m_total_counts.num_files;
}


 /*
 void Line_Counter::count( const std::string& file_name )
{
   ifstream in( file_name, std::ios::in );
   if( !in )
   {
      const auto msg { string( "Couldn\'t open file " ) + file_name };
      throw runtime_error( msg.c_str() );
   }
   m_file_counts.clear();
   string line;
   bool in_comment { false };
   while( getline( in, line ) )
   {
      line = trim( line );

      // all is_*_line() require a trimmed line for input
      if( in_comment || is_comment_start_line( line ) )
      {
         in_comment = true; // may not be set if start line
         ++m_file_counts.num_long_comments;
         if( is_comment_end_line(line) )
            in_comment = false;
      }

      // not in a long comment for rest of these
      else if( is_blank_line( line ) )
         ++m_file_counts.num_blanks;

      else if( is_comment_line( line ) )
         ++m_file_counts.num_short_comments;

      else if( is_class_line( line ) )
         ++m_file_counts.num_classes;

      else if( is_struct_line( line ) )
         ++m_file_counts.num_structs;

      else // it's gotta be code
         ++m_file_counts.num_code;
   } // end of while loop

   // didn't count structs and classes as lines of code
   // so do that now
   m_file_counts.num_code +=
      m_file_counts.num_classes + m_file_counts.num_structs;

   // now add stuff for this file into totals
   m_total_counts += m_file_counts;

   ++m_total_counts.num_files;
}
*/