#ifndef LINE_COUNTER_HPP
#define LINE_COUNTER_HPP

#include <string>
#include <vector>

namespace FJ
{
   /* Line_Counter
   * This class counts the number of lines of different types in a
   * text file that presumably holds C or C++ source code. The
   * 
   * Terminology:
   * 
   * blank line - a line that is empty or contains only whitespace characters
   * code line - a line that is not a blank line, a short comment or part of
   *     a long comment
   * long comment - text that starts on a line whose first two non-whitespace
   *     characters are "/*" and ends on the next line whose last two
   *     non-whitespace characters are "*" followed immediately by "/".
   *     (If I write those two characters together the editor thinks that 
   *     terminates this comment!) The two terminating characters
   *     can be on the same line as the starting characters or a different line.
   * short comment - a line whose first two non-whitespace characters are "//"
   *     and that does not appear within a long comment.
   */
   class Line_Counter
   {
   public:
      Line_Counter( std::string file_name = "" );
      // constructor
      // if file_name is not empty, calls count() on that file
      // if there's a problem, throws std::runtime_error

      Line_Counter( const Line_Counter& source ) = default;
      // default copy constructor

      Line_Counter& operator=( const Line_Counter& rhs ) = default;
      // default copy assignment

      Line_Counter( Line_Counter&& source ) = default;
      // default move constructor

      Line_Counter& operator=( Line_Counter&& rhs ) = default;
      // default move assignment

      void clear_totals();
      // clears all total counts (but not file_counts)

      void count( const std::string& file_name );
      // count the lines in the file and add those numbers to the totals

      int num_blank_lines() const;
      // number of blank lines in last file proccesed
      // does not include blank lines within long comments

      int num_classes() const;
      // number of classes in last file proccesed

      int num_code_lines() const;
      // number of lines of code in last file proccesed

      int num_comment_lines() const;
      // sum of number of short comment lines,
      // and long comment lines in last file processed

      int num_lines() const;
      // total number of lines in last file processed

      int num_long_comment_lines() const;
      // number of long comment lines in last file proccesed
      // includes blank lines and short comment lines inside
      // a long comment

      int num_noncode_lines() const;
      // sum of blank lines and comment lines

      int num_short_comment_lines() const;
      // number of short comment lines in last file proccesed
      // does not include short comment lines inside long comments

      int num_structs() const;
      // number of structs in last file proccesed

      float ratio_code_noncode() const;
      // number of code lines divided by number of noncode lines
      // if the latter is zero, returns a negative number

      int total_blank_lines() const;
      // total number of blank lines
      // does not include blank lines within long comments

      int total_classes() const;
      // total number of classes in all files

      int total_code_lines() const;
      // number of lines of code

      int total_comment_lines() const;
      // sum of number of blank lines, short comment lines,
      // and long comment lines in all files

      int total_files() const;
      // total number of files
         
      int total_lines() const;
      // total number of lines in all files
         
      int total_long_comment_lines() const;
      // total number of long comment lines in all files

      int total_noncode_lines() const;
      // sum of total blank lines and total comment lines

      float total_ratio_code_noncode() const;
      // total number of code lines divided by total number of noncode lines
      // if the latter is zero, returns a negative number

      int total_short_comment_lines() const;
      // total number of short comment lines in all files

      int total_structs() const;
      // total number of structs in all files

   protected:
      void count( const std::vector<std::string>& lines );


   private:

      // IMPORTANT - all of the is_*_line() methods assume the input
      // line has been trimmed, i.e. all leading and trailing white
      // space removed

      bool is_blank_line( const std::string& line ) const;
      // return true iff line is empty

      bool is_class_line( const std::string& line ) const;
      // return true iff the first word is "class" and there is no
      // semicolon in the line. (This prevents a forward declaration
      // from being counted as a class line).

      bool is_comment_end_line( const std::string& line ) const;
      // return true iff the last two characters are "*/"

      bool is_comment_line( const std::string& line ) const;
      // return true iff the first two characters are "//"

      bool is_comment_start_line( const std::string& line ) const;
      // return true iff the first two characters are "/*"

      bool is_struct_line( const std::string& line ) const;
      // return true iff the first word is "struct" and there is no
      // semicolon in the line. (This prevents a forward declaration
      // from being counted as a class line).

      struct Counts
      {
         void operator+=( const Counts& rhs )
         {
            num_blanks += rhs.num_blanks;
            num_classes += rhs.num_classes;
            num_code += rhs.num_code;
            num_files += rhs.num_files;
            num_long_comments += rhs.num_long_comments;
            num_short_comments += rhs.num_short_comments;
            num_structs += rhs.num_structs;
         }
         void clear() 
         {
            num_blanks = num_code = num_long_comments = num_short_comments 
               = num_classes = num_structs = num_files = 0;
         };
         int total() const 
         { return num_blanks + num_code + num_long_comments + num_short_comments; }
         int num_blanks { 0 };
         int num_classes { 0 };
         int num_code { 0 };
         int num_files { 0 };
         int num_long_comments { 0 };
         int num_short_comments { 0 };
         int num_structs { 0 };
        
      };
      // internal structure to hold number of times various items occurred

      Counts m_file_counts; 
      Counts m_total_counts;

      static constexpr float m_invalid { -1.0f };
   };

} // end namespace gjr
#endif