#include "Settings.hpp"
#include "utilities.hpp"

#include "DOCUMENT.H"

#include "rapidjson_utilities.hpp"

#include <filesystem>
#include <fstream>
#include <iostream>
#include <sstream>

using namespace FJ;
namespace fs = std::filesystem;
using namespace rapidjson;
using namespace std;

Settings::Settings()
{
   mExecutablePath = executablePath();
}

Settings::Settings( const std::vector<NameValue>& settings, const std::vector<InputFile>& inputFiles, 
   const std::vector<OutputFile>& outputFiles, const std::vector<std::string>& calculationNames, 
   const std::string& summaryFile )
   : Settings()
{
   store( settings,inputFiles,outputFiles,calculationNames, summaryFile );
}

std::string Settings::addPath( std::string fileName ) const
{
   if( !mExecutablePath.empty() )
   {
      fs::path temp( mExecutablePath );
      temp /= fileName;
      fileName = temp.u8string(); // native format
   }
   return fileName;
}

std::vector<std::string> Settings::calculationNames() const
{
   return mCalculationNames;
}

std::string Settings::executablePath() const
{
   return GetExecutablePath();
}

bool Settings::has_setting( const std::string& key ) const
{
   return find_if( mSettings.cbegin(), mSettings.cend(),
      [&key]( const auto& nameValue ) {return nameValue.name()==key; } )
   != mSettings.cend();
}

NameValue Settings::setting( const std::string& key ) const
{
   const auto itr = find_if( mSettings.cbegin(),mSettings.cend(),
      [&key]( const auto& nameValue ) {return nameValue.name()==key; } );
   if( itr != mSettings.cend() )
      return *itr;
   else
   {
      ostringstream out;
      out << "There is no setting whose name is \"" << key << "\"";
      throw runtime_error( out.str() );
   }
}

Settings::InputFile Settings::inputFile( int index, bool addThePath ) const
{
   InputFile file( mInputFiles[index] );
   if( addThePath )
      file.name = addPath( file.name );
   return file;
}

std::vector<Settings::InputFile> Settings::inputFiles() const
{
   return mInputFiles;
}

int Settings::numInputFiles() const
{
   return static_cast<int>(mInputFiles.size()); 
}

int Settings::numOutputFiles() const
{
   return static_cast<int>(mOutputFiles.size()); 
}

Settings::OutputFile Settings::outputFile( int index, bool addThePath ) const
{
   OutputFile file( mOutputFiles[index] );
   if( addThePath )
      file.name = addPath( file.name );
   return file;
}

std::vector<Settings::OutputFile> Settings::outputFiles() const
{
   return mOutputFiles;
}

std::vector<NameValue> Settings::settings() const
{
   return mSettings;
}

std::string Settings::summaryFile( bool addThePath ) const
{
   string file( mSummaryFile );
   if( addThePath )
      file = addPath( file );
   return file;
}

//void Settings::add_calculations( const std::vector<Calculation>& results )
//{
//   mCalculations.insert( mCalculations.begin(),results.cbegin(),results.cend() );
//} 
//
//void Settings::add_calculation( const Calculation& result )
//{
//   mCalculations.push_back( result );
//}
//
//void Settings::clear_calculations()
//{
//   mCalculations.clear();
//}
//
//std::string Settings::information() const
//{
//   return mInformation;
//}
//
//void Settings::information( const std::string& info )
//{
//   mInformation = info;
//}
//
//bool Settings::is_success() const
//{
//   return m_success;
//}
//
//void Settings::is_success( bool success )
//{
//   m_success = success;
//}

//void Settings::read_file( const std::string& fileName )
//{
//   try
//   {
//      ifstream file( fileName.c_str() );
//      if( !file )
//         throw runtime_error( "Couldn\'t open JSON settings file" );
//
//      string contents( (std::istreambuf_iterator<char>( file )), // extra parentheses necessary
//         std::istreambuf_iterator<char>() );
//
//      Document doc;
//      
//
//   }
//   catch( exception& e )
//   {
//
//   }
//
//}

void Settings::store( const std::vector<NameValue>& settings, const std::vector<InputFile>& inputFiles,
   const std::vector<OutputFile>& outputFiles, const std::vector<std::string>& calculationNames, 
   const std::string& summaryFile )
{

   mInputFiles = inputFiles;
   mSettings = settings;
   mOutputFiles = outputFiles;
   mCalculationNames = calculationNames;
   mSummaryFile = summaryFile;
}

void Settings::processLoadedJSON()
{}

void Settings::loadJSONSetup( const std::string& settingsFile )
{
   doLoadJSONSetup( settingsFile );
   processLoadedJSON();
}

void Settings::doLoadJSONSetup( const std::string& settingsFile )
{
   //const auto cp = fs::current_path().u8string();
   ostringstream text;
   ifstream in( settingsFile.c_str() );
   if( !in )
   {
      text << "Couldn\'t open " << settingsFile;
      throw runtime_error( text.str() );
   }

   try
   {
      Document doc;
      auto& a = doc.GetAllocator();

      string line;
      ostringstream content;
      while( getline(in,line) )
         content << line;

      in.close();

      doc.Parse( content.str().c_str() );
      auto temp = prettyJSON( doc );
      clear();

      // read optional names of input files. Usually present and with one name
      if( doc.HasMember( inputFiles_Att ) )
      {
         const auto files = doc[inputFiles_Att].GetArray();
         Settings::InputFile inputFile;
         for( int i = 0; i < files.Size(); ++i )
         {
            const auto& temp = files[i];
            if( temp.HasMember( name_Att ) )
               inputFile.name = temp[name_Att].GetString();
            else
            {
               text << "Input file "    << (i+1) << " has no name";
               throw runtime_error( text.str() );
            }
            inputFile.is_normalized = 
               temp.HasMember( isNormalized_Att ) ? temp[isNormalized_Att].GetBool() : defaultIsNormalized;
            mInputFiles.push_back( inputFile );
         }
      }

      // read optional settings. Usually present and with multiple entries
      if( doc.HasMember( settings_Att ) )
      {
         int ix( 0 );
         for( auto itr = doc[settings_Att].MemberBegin(); itr!=doc[settings_Att].MemberEnd(); ++itr,++ix )
            mSettings.push_back( fromJSON( itr,ix ) );

      }

      // read required name of summary file
      if( doc.HasMember( summaryFile_Att ) && !doc[summaryFile_Att].IsNull() )
      {
         mSummaryFile = doc[summaryFile_Att].GetString();
      }
      else
      {
         text << "The summary-file name is missing";
         throw runtime_error( text.str() );
      }

      // read optional names of output files. Usually present and with one name
      if( doc.HasMember( outputFiles_Att ) )
      {
         const auto files = doc[outputFiles_Att].GetArray();
         Settings::OutputFile outputFile;
         mOutputFiles.reserve( files.Size() );
         for( int i = 0; i < files.Size(); ++i )
         {
            const auto& temp = files[i];
            if( temp.HasMember( name_Att ) )
               outputFile.name = temp[name_Att].GetString();
            else
            {
               text << "Output file "    << (i+1) << " has no name";
               throw runtime_error( text.str() );
            }

            if( temp.HasMember( outputColumnLabels_Att ) )
            {
               if( !temp[outputColumnLabels_Att].IsArray() )
               {
                  text << "Output file "    << (i+1) << " column labels must be in a JSON array";
                  throw runtime_error( text.str() );
               }
               const auto& labels = temp[outputColumnLabels_Att].GetArray();
               outputFile.columnLabels.reserve( labels.Size() );
               for( const auto& label : labels )
                  if( label.IsString() )
                     outputFile.columnLabels.push_back( label.GetString() );
                  else
                  {
                     text << "Output file "    << (i+1) << " has a column label that is not text";
                     throw runtime_error( text.str() );
                  }
            }
            mOutputFiles.push_back( outputFile );
         }
      }
   }
   catch( runtime_error& )
   {
      throw;
   }
   catch( exception& error )
   {
      throw runtime_error( error.what() );
   }
   catch(...)
   {
      text << "Couldn\'t parse JSON in " << settingsFile;
      throw runtime_error( text.str() );
   }
}

NameValue Settings::fromJSON( const rapidjson::Value::MemberIterator& itr, int ix ) const
{
   ostringstream text;
   if( itr->name.IsNull() )
   {
      text << "NameValue " << (ix+1) << " must have a name (key)";
      throw runtime_error( text.str() );
   }

   if( itr->value.IsNull() )
   {
      text << "Value " << (ix+1) << " is missing or empty";
      throw runtime_error( text.str() );
   }

   const string name { itr->name.GetString() };
   if( itr->value.IsBool() )
      return NameValue( name,itr->value.GetBool() );
   else if( itr->value.IsInt() )
      return NameValue( name,itr->value.GetInt() );
   else if( itr->value.IsDouble() )
      return NameValue( name,itr->value.GetDouble() );
   else if( itr->value.IsString() )
      return NameValue( name,itr->value.GetString() );
   else
   {
      text << "Value must be a boolean, integer, floating-point or string";
      throw runtime_error( text.str() );
   }
}

void Settings::clear()
{
   mCalculationNames.clear();
   //mExecutablePath.clear(); don't clear this because it's not something that's read in
   mInputFiles.clear();
   mOutputFiles.clear();
   mSettings.clear();
   mSummaryFile.clear();
}