#ifndef SETTINGS_HPP
#define SETTINGS_HPP

#include "DOCUMENT.H"

#include "NameValue.hpp"

#include <vector>

/*
Settings reads general setup (settings) information for a plugin. Currently it does
this only by reading a JSON file. Users should derive from this class to store
the input as data members for convenience, validate the input, etc.
See processLoadedJSON()

Currently all input and output files are assumed to be in the folder that
contains the C++ executable so file names should never contain a path.
An example of the general format is:
{
   "input_files" :
   [
      {
         "name" : "file1.csv",
         "is_normalized" : true
      }
   ],
   "settings" :
   {
      "setting1" : "complete",
      "setting2" : 12,
      "setting3" : 10.0,
      "setting4" : true
   },
   "summary_file" : "summary.json",
   "output_files" :
   [
      {
         "name" : "output1.csv",
         "column_labels" : [ "X", "Y", "Z" ]
      }
   ]
}

input_files
- optional, though usually present. Zero or more members. In a member:
   - name - required, string. Name of an input file with no path
   - is_normalized - optional, boolean, default is true. Whether or not the data in the file is normalized

settings
- optional, though usually present. Zero or more name-value pairs. In a pair:
   - name - required, string. Name of setting
   - value - required, bool, int, float or string. Value of setting

- summary_file
- required, string. Name of file in which the plugin places a summary of the run. 
  See the class Summary

- output_files
- optional, though usually present. Zero or more members. In a member:
- name - required, string. Name of an input file with no path
- is_normalized - optional, boolean, default is true. Whether or not the data in the file is normalized
- column_labels - optional. Array of column labels to be used in output CSV file, starting with first column

*/
namespace FJ
{
   class Settings
   {
   public:

      struct InputFile
      {
         std::string name;
         bool is_normalized { true };
      };

      struct OutputFile
      {
         std::string name;
         std::vector<std::string> columnLabels;
      };

      using Calculation = NameValue;
   public:

      Settings();
      Settings( const std::vector<NameValue>& settings, const std::vector<InputFile>& inputFiles, 
          const std::vector<OutputFile>& outputFiles, const std::vector<std::string>& calculation_names, 
         const std::string& summaryFile );

      Settings( const Settings& source ) = default;
      Settings& operator=( const Settings& rhs ) = default;

      Settings( Settings&& source ) = default;
      Settings& operator=( Settings&& rhs ) = default;

      virtual ~Settings() = default;

      std::vector<std::string> calculationNames() const;

      bool has_setting( const std::string& key ) const;
      // return true if setting with given key exists (case-sensitive), else false

      Settings::InputFile inputFile( int index, bool addPath=true ) const;
         
      std::vector<InputFile> inputFiles() const;

      void loadJSONSetup( const std::string& settingsFile );
         
      int numInputFiles() const;

      int numOutputFiles() const;

      OutputFile outputFile( int index, bool addPath=true ) const;
      // assumes numOutputFiles()>0 and 0 <= index < numOutputFiles()

      std::vector<OutputFile> outputFiles() const;

      NameValue setting( const std::string& name ) const;
      // return NameValue of setting with given key (case-sensitive)
      // If not there, throws std::runtime_error

      std::vector<NameValue> settings() const;

      std::string summaryFile( bool addPath=true) const;

      //void add_calculations( const std::vector<Calculation>& results );
      //void add_calculation( const Calculation& result );
      //void clear_calculations();

   protected:

      std::string addPath( std::string fileName ) const;

      void clear();
      // clears all member data

      std::string executablePath() const;

      NameValue fromJSON( const rapidjson::Value::MemberIterator& itr, int ix ) const;

      virtual void processLoadedJSON();
      // Called automatically after doLoadJSONSetup and by default does nothing. 
      // Derived classes should overwrite to store loaded data
      // in derived members, validate data, etc.

   private:

      inline static const char* inputFiles_Att { "input_files" };
      inline static const char* isNormalized_Att { "is_normalized" };
      inline static const char* name_Att { "name" };
      inline static const char* outputColumnLabels_Att { "column_labels" };
      inline static const char* outputFiles_Att { "output_files" };
      inline static const char* settings_Att { "settings" };
      inline static const char* summaryFile_Att { "summary_file" };
      inline static const bool defaultIsNormalized { true };

      //void read_file( const std::string& fileName );

      void doLoadJSONSetup( const std::string& settingsFile );

      void store( const std::vector<NameValue>& settings, const std::vector<InputFile>& inputFiles,
         const std::vector<OutputFile>& outputFiles, const std::vector<std::string>& calculationNames, 
         const std::string& summaryFile  );

      std::vector<std::string> mCalculationNames;

      std::string mExecutablePath;

      std::vector<InputFile> mInputFiles;

      std::vector<OutputFile> mOutputFiles;

      std::vector<NameValue> mSettings;

      std::string mSummaryFile;
   };

} // end namespace FJ
#endif