#include "flowjo_utilities.hpp"

#include "flowjo_exceptions.hpp"

#include <filesystem>
#include <fstream>
#include <regex>

using namespace FJ;
using namespace std;
using namespace std::filesystem;

int FJ::case_insensitive_comparison( const std::string& s1, 
   const std::string& s2 )
{
   auto less_than_insensitive = []( char c1, char c2 )
   {
      return  tolower( static_cast<unsigned char>(c1) )
         < tolower( static_cast<unsigned char>(c2) );
   };
   // return true if c1 < c2, false otherwise

   if( lexicographical_compare( s1.begin(), s1.end(),
      s2.begin(), s2.end(), less_than_insensitive ) )
      return -1;
   if( lexicographical_compare( s2.begin(), s2.end(),
      s1.begin(), s1.end(), less_than_insensitive ) )
      return 1;
   return 0;
}
// from "C++ Standard Library Practical Tips", by Greg Reese

int FJ::compare( const std::string s1, const std::string s2,
   bool case_sensitive )
{
   return case_sensitive ? 
      s1.compare( s2 ) : case_insensitive_comparison( s1, s2 );
}

void FJ::count_lines( Line_Counter& counter, const std::string& file_name,
   const std::vector<std::string>& extensions )
{

   try
   {
      path file_path( file_name );
      string extension = file_path.extension().string();

      // return extension (including leading period) or blank if file has no extension
      if( !extension.empty() )
         extension.erase( 0, 1 ); // erase leading char (.)
      if( find_if( extensions.cbegin(), extensions.cend(),
         [&extension]( const auto& x ) {return case_insensitive_comparison( extension, x ) == 0; } )
         != extensions.cend() )
      {
         counter.count( file_name );
      }
   }
   catch( ... )
   {
      return; // do nothing on error
   }
   return;
}

void FJ::count_lines( Line_Counter& counter, const std::string& path_text,
   const std::vector<std::string>& extensions, bool recursive )
{
   try
   {
      auto p = std::filesystem::current_path();
      if( recursive )
      {
         path file_path( path_text );
         recursive_directory_iterator itr { file_path };
         for( const auto& file : itr ) // file is a directory_entry
            if( file.is_regular_file() )
               count_lines( counter, file.path().string(), extensions );
      }
      else // not recursive
      {
         path file_path( path_text );
         for( const auto& file : directory_iterator( file_path ) )// file is a directory_entry
            if( file.is_regular_file() )
               count_lines( counter, file.path().string(), extensions );
      }
   }
   catch( const filesystem_error& error )
   {
      auto code = error.code(); // for debugging
      string text = error.what(); // for debugging
   }
   catch( ... )
   {
      return;
   }
}

Line_Counter FJ::count_lines( const std::string& path_text,
   const std::vector<std::string>& extensions, bool recursive )
{
   Line_Counter counter;
   count_lines( counter, path_text, extensions, recursive );
   return counter;
}

std::string::size_type FJ::find_text( const std::string& file_name,
   const std::string& text )
{
   vector<string> texts( 1, text );
   const auto positions = find_text( file_name, texts );
   return positions.front();
}

std::vector<std::string::size_type> FJ::find_text( const std::string& file_name,
   const std::vector<std::string>& strings )
{
   ifstream in( file_name, ios::binary );
   if( !in )
      throw Bad_File_Open();
   string file_text;

   // find out how many characters we're going to store and resize
   // the string to avoid having it to resize repeatedly while reading
   // in the data and thus hopefully speeding things up
   in.seekg( 0, ios_base::end ); // go to end of file
   const auto file_size = in.tellg(); // get size
   in.seekg( 0 ); // rewind to begining
   file_text.resize( static_cast<string::size_type>(file_size) ); // assume cast okay

   // read file directly into string
   // CAUTION - do not use istream_iterator because by default, it skips whitespace
   // Use istreambuf_iterator which doesn't skip whitespace and is more efficient 
   file_text.assign( istreambuf_iterator<char>( in ), istreambuf_iterator<char>() );

   vector<string::size_type> positions; // need size_type because could return string::npos
   positions.reserve( strings.size() );
   for( const auto& s : strings )
      positions.push_back( file_text.find( s ) ); // returns string::npos if can't find

   return positions; // assume can cast
}

std::vector<std::string> FJ::make_lines( const std::string& file )
{
   ifstream in( file, ios::binary | ios::ate ); // ate means go to end of stream immediately after opening
   if( !in )
      throw Bad_File_Open( file );

   // read the whole file, including line terminators, into one string
   const auto size = in.tellg();
   string text( size, '\0' );
   in.seekg( 0 );
   if( !in.read( &text[0], size ) )
      throw Bad_File_Read();

   vector<string> lines;
   // if empty file, return empty vector of lines
   if( size == 0 )
      return lines;

   // at this point know that there is at least one character in the file
   static const char CR { 13 }; // carriage return
   static const char LF { 10 }; // line feed

   // to start, assume terminator is CR 
   char terminator { CR };    // first, and maybe only, character of terminator
   int terminator_len { 1 };  // length (number of characters) in terminator

   // look for a terminator
   auto ix = text.find( CR );
   const auto text_size = text.size();

   // if we found a CR
   if( ix != string::npos )
   {
      if( ix < text_size - 1 && text[ix + 1] == LF )
         terminator_len = 2;
   }

   // look for a LF
   else
   {
      ix = text.find( LF );

      // if we found a LF
      if( ix != string::npos )
      {
         terminator = LF;
      }

      // there was no CR and no LF and the file has at least one character.
      // This means the entire file is one line. Store it and return
      else
      {
         lines.emplace_back( text );
         return lines;
      }
   }

   // at this point we know that text is not empty and we know the
   // terminator and its length
   auto itr1 = text.cbegin();
   auto itr2 = itr1 + ix;
   while( true )
   {
      lines.emplace_back( itr1, itr2 );

      // When Visual Studio is in debug mode, the sum on the right in the line
      // itr1 = itr2 + terminator_len;
      // causes an exception to be thrown if the sum would be > text.cend(). 
      // When in release mode, VS doesn't throw
      // Work around this
#if defined(WINDOWS) && defined(_DEBUG)
      if( itr2==text.cend() || (terminator_len==2 && itr2==text.cend()-1) )
         break;
#endif
      itr1 = itr2 + terminator_len;
      if( itr1 >= text.cend() )
         break;

      itr2 = find( itr1, text.cend(), terminator );
   };
   return lines;
}

void FJ::replace_all( std::string& text, char target, char replacement )
{
   for( auto& c : text )
      if( c == target )
         c = replacement;
}

std::string::size_type 
FJ::replace_all( std::string& text, const std::string& target,
   const string& replacement )
{
   // algorithm works even if target==replacment
   const auto target_len = target.length();
   const auto replacement_len = replacement.length();
   std::string::size_type ix { 0 };
   std::string::size_type num_replacements { 0 };
   while( (ix = text.find( target, ix )) != string::npos )
   {
      text.replace( ix, target_len, replacement );
      ix += replacement_len;
      ++num_replacements;
   }
   return num_replacements;
}

std::vector<std::string> FJ::tokenize( const std::string& text,
   const std::string& delimiters, bool allow_escaped_delimiters )
{
   vector<string> tokens;

   // if no text, no tokens
   if( text.empty() )
      return tokens;

   // if no delimiters, entire text is one token
   if( delimiters.empty() )
   {
      tokens.push_back( text );
      return tokens;
   }

   // at this point know that text and delimiters each have at least
   // one character
   const auto text_size = text.size();
   string token;
   for( string::size_type ix = 0; ix < text_size; ++ix )
   {
      // store the current character for convenience
      const auto current_char = text[ix];

      // see if the current character is a delimiter
      const auto is_delimiter = 
         delimiters.find( current_char ) != string::npos;
      
      // if the char is a delimiter
      if( is_delimiter )
      {
         if( allow_escaped_delimiters )
         {
            // if this is the last character it can't be escaped
            // so this is the end of the token. The loop exits
            // at the end of this iteration and the code that 
            // follows will store the token if it's not empty

           // if there's at least one more character of text 
            if( ix < text_size - 1 )
            {
               // if the next character is the same as the
               // current one, i.e., we have two of the same
               // delimiters in a row, add the character to 
               // the token and skip the next character
               if( text[ix+1] == current_char )
               {
                  token += current_char;
                  ++ix;
               }

               // the next character is not the same as the
               // current one so this is the end of the token
               else
               {
                  if( !token.empty() )
                  {
                     tokens.push_back( token );
                     token.clear();
                  }
               }
            }
         }

         // if we're not allowing escaped characters this is
         // the end of the token. If it's not empty, save it 
         // for output and clear it to be ready for a new token
         else
         {
            if( !token.empty() )
            {
               tokens.push_back( token );
               token.clear();
            }
         }
      }
      // if the char is not a delimiter
      else
         token += text[ix];
   } // end of loop

   if( !token.empty() )
      tokens.push_back( token );

   return tokens;
}

std::string FJ::trim( const std::string& text )
{
   if( text.empty() ||
      // the first and last chars are both not whitespace
      (!isspace( static_cast<unsigned char>(text.front()) )
         && !isspace( static_cast<unsigned char>(text.back()) )
         ) ) 
      return text;
   else
   {
      string trimmed; // empty string
      string::size_type start_ix = 0;

      // find first non-whitespace
      for( ; start_ix < text.length(); ++start_ix )
         if( !isspace( static_cast<unsigned char>(text[start_ix]) ) )
            break;

      // if we didn't find a non-whitespace return an empty string
      if( start_ix == text.length() )
         return trimmed;

      // otherwise, find the last non-white
      else
      {
         string::size_type end_ix = text.length() - 1;
         for( ; end_ix >= start_ix; --end_ix )
            if( !isspace( static_cast<unsigned char>(text[end_ix]) ) )
               break;
         auto count = end_ix - start_ix + 1;

         // range is [start_ix,start_ix+count), 
         // i.e. doesn't include last element in range
         trimmed = text.substr( start_ix, count );
         return trimmed;
      }
   }
}

std::string FJ::to_upper( std::string text )
{
   transform( text.begin(), text.end(), text.begin(),
      []( unsigned char c ) { return std::toupper( c ); } // correct
   );
   // taken from https://en.cppreference.com/w/cpp/string/byte/toupper
   return text;
}

std::vector<std::string> FJ::regex_file_match( std::string user_filename,
   std::string user_path, bool case_sensitive )
{
   // if the user didn't specify a file name, match everything
   if( user_filename.empty() )
      user_filename = R"(.*)";

   // if user didn't specify a path, put in a period, which means to search
   // in the current directory. Note that filesystem_path is not a regular expression
   // so you don't need to escape the period
   filesystem::path filesystem_path { user_path.empty() ? "." : user_path };

   regex_constants::syntax_option_type flags { regex::ECMAScript }; // this is default
   if( !case_sensitive )
      flags |= regex::icase;

   // CAUTION - constructor will throw std::regex_error if text isn't valid regular expression
   regex regex_filename( user_filename, flags );

   // to search subdirectories also, use recursive_directory_iterator
   vector<string> files;

   // CAUTION - if filesystem_path is bad, directory_iterator throws std::filesystem::filesystem_error
   for( const auto& entry :
      filesystem::directory_iterator( filesystem_path, directory_options::skip_permission_denied ) )
   {
      // skip_permission_denied causes directory_iterator to skip DIRECTORIES to which the code would
      // not have access. However, it does not work on files. The try-catch block just checks if the
      // file exist but doesn't use the result. The purpose of checking is that the code will throw an
      // exception is access is denied. If so, the code ignores the file and moves on to the next.
      // https://stackoverflow.com/questions/66559208/stdfilesystem-recursive-iterator-will-throw-permission-denied-even-when-skip-p
      try
      {
         const auto dummy = exists(entry); // store result to quiet compiler
      }
      catch( ... )
      {
         continue;
      }
      // If it's a regular file
      if( is_regular_file( entry )         
         // and the stem and extension match, save it 
         && regex_match( entry.path().filename().u8string(), regex_filename )
         )
         files.push_back( entry.path().u8string() );
   }

   return files;
}

std::vector<std::string> FJ::wildcard_file_match( std::string filename, 
   std::string path, bool case_sensitive )
{
   // a period in a regular expression denotes any character. Escape it to
   // make it represent a literal period
   replace_all( filename, ".", "\\." );

   // an asterisk in a wildcard specification means match zero or more characters
   // Replace it with ".*", its regex equivalent
   replace_all( filename, "*", ".*" );

   // a question mark in a wildcard specification means match exactly one character
   // Replace it with ".{1}", its regex equivalent
   replace_all( filename, "?", ".{1}" );

   return regex_file_match( filename, path, case_sensitive );
}

std::vector<std::string> FJ::wildcard_file_match( std::string full_path,
   bool case_sensitive )
{
   path the_path( full_path );
   auto directory = the_path.parent_path().u8string(); // includes drive
   auto filename = the_path.filename().u8string(); // stem and extension
   return wildcard_file_match( filename, directory, case_sensitive );
}