#ifndef FLOWJO_UTILITIES_HPP
#define FLOWJO_UTILITIES_HPP

#include "Line_Counter.hpp"

namespace FJ
{
   /* General purpose C++ utilities. Must compile with at least C++14
   *  Created by Greg Reese on February 15, 2022
   */
   int case_insensitive_comparison( const std::string& s1, const std::string& s2 );
   // case-insensitive comparison
   // INPUTS
   // s1 - first string
   // s2 - second string
   // RETURN - return < 0 if s1 < s2, return > 0 if s1 > s2, return 0 if s1 == s2
   // NOTES 
   // 1) either or both strings can be empty or any size

   int compare( const std::string s1, const std::string s2,
      bool case_sensitive=true );
   // string comparison with specified case sensitivity
   // INPUTS
   // s1 - first string
   // s2 - second string
   // RETURN - return < 0 if s1 < s2, return > 0 if s1 > s2, return 0 if s1 == s2
   // NOTES 
   // 1) either or both strings can be empty or any size

   Line_Counter count_lines( const std::string& path,
      const std::vector<std::string>& extensions, bool recursive );
   // gathers line statistics of all files with specified extensions
   // 
   // Use this function to examine all C++ files, e.g. *.cpp and *.hpp,
   // on a path and get back the group statistics. See the documentation
   // in Line_Counter.hpp for the statistics gathered.
   // 
   // INPUTS
   // path - path to folder containing files to be processed. path can have a 
   //    terminating separator or not
   // extensions - file extensions to consider (no period in extensions)
   // recursive - if true, processes folder and recursively all subfolders
   //             if false, processes only folder
   // RETURN
   // A Line_Counter with total statistics of all files that
   // ERRORS
   // If there are problems opening or reading a file, the routine
   // ignores it
   // 
   // NOTES
   // 1) comparison of extensions is case-insensitive
   // 2) an empty string for an extension is legal and means that the file has
   //    no extension

   void count_lines( Line_Counter& counter, const std::string& file_name,
      const std::vector<std::string>& extensions );
   // gathers line statistics in a file if file has one of specified extensions
   // INPUTS
   // counter - if file has desired extension, resets local statistics in counter
   //    (but not total statistics), runs counter. If file doesn't have desired
   //    extension, does nothing
   // file_name - name of file to count lines in
   // extensions - file extensions to consider (no period in extensions)
   // ERRORS
   // If there are problems finding the extension or opening the file, the routine
   // returns without doing anything
   // 
   // NOTES
   // 1) comparison of extensions is case-insensitive
   // 2) an empty string for an extension is legal and means that the file has
   //    no extension

   void count_lines( Line_Counter& counter, const std::string& path,
      const std::vector<std::string>& extensions, bool recursive );
   // gathers line statistics on all files on path that have specified extensions
   // INPUTS
   // counter - if file has desired extension, resets local statistics in counter
   //    (but not total statistics), runs counter. If file doesn't have desired
   //    extension, does nothing
   // path - path to folder containing files to be processed. path can have a 
   //    terminating separator or not
   // extensions - file extensions to consider (no period in extensions)
   // recursive - if true, processes folder and recursively all subfolders
   //             if false, processes only folder
   // ERRORS
   // If there are problems finding the extension or opening a file, the routine
   // ignores it
   // 
   // NOTES
   // 1) comparison of extensions is case-insensitive
   // 2) an empty string for an extension is legal and means that the file has
   //    no extension

   std::string::size_type find_text( const std::string& file_name,
      const std::string& text );
   // search for text in a file
   // INPUTS
   // file_name - name of file to open and search (can have path)
   // text - text to look for
   // RETURN - if text found, return is index (zero-based byte number) of start
   // of text in file. If text not found, return std::string::npos
   // NOTES 
   // 1) Function works for binary and text files
   // 2) Match is always case-sensitive
   // 3) If can't open file, throws FJ::Bad_File_Open

   std::vector<std::string::size_type> find_text( const std::string& file_name,
      const std::vector<std::string>& strings );
   // search for multiple text strings in a file
   // INPUTS
   // file_name - name of file to open and search (can have path)
   // strings - text strings to look for
   // RETURN - zero-based index of first instance found of each text string.
   // If text found not found, return std::string::npos
   // NOTES 
   // 1) Function works for binary and text files
   // 2) Match is always case-sensitive
   // 3) If can't open file, throws FJ::Bad_File_Open

   std::vector<std::string> make_lines( const std::string& file );
   // read lines of text file whether terminated by CRLF, CR or LF
   // INPUTS
   // file - file name
   // RETURN - vector of lines in file without terminator
   // NOTES 
   // 1) an empty string means the line consisted only of a terminator
   // 2) Terminators
   //    Windows - CRLF (carriage return, line feed)
   //    classic MacOS - CR
   //    MacOS - LF

   std::vector<std::string> regex_file_match( std::string filename = std::string(),
      std::string path = std::string(), bool case_sensitive=false );
   // Get file names that match a regular expression
   // INPUTS
   // filename - the part of the file specification after the directory
   //   e.g., sp97.jo in c:\greg\sp97.jo    
   //   You can use regular expressions anywhere in the file name. To put in a period,
   //   use \.    See Note 1) below
   // path - No regular expressions, can be relative or absolute, can have or omit final directory separator
   //   Forward slash (/) works for both Mac and Windows, back slash (\) works for Windows only
   //   Leave empty or use period to search in current directory
   // case_sensitive - if true match is case sensitive, else case insensitive
   // RETURN - vector of file names with absolute paths
   // NOTES 
   // 1) use common file-name patterns by putting the following within the parentheses of R"()"
   // i.e., a raw string
   // .*                       all file names
   // [^\.]*                   file names without an extension, e.g., "engine_version"
   // \.[^\.]+                 file names with no stem but a non-empty extension, e.g., ".gitignore"
   // test\.[^\.]+             all files with a given case-insensitive stem, e.g., "test", and a non-empty extension
   // s[^\.]*\.[^\.]+          all files with a stem starting with an "s" and a non-empty extension, e.g. "sp9cdc.jo"
   // s[^\.]*\d\d\.[^\.]+      all files with a stem starting with an "s" and ending with at least two digits 
   //                            and having a non-empty extension, e.g. "spa97.jo"
   // 2) calling the function with no arguments does a case-insensitive match to all files in the
   //    current directory

   void replace_all( std::string& text, char target, char replacement );
   // replace all occurrences of a character with another
   // INPUTS
   // text - text to replace in
   // target - character to be replaced
   // replacement - character substituted for target character
   // NOTES 
   // 1) Can replace character by itself

   std::string::size_type
   replace_all( std::string& text, const std::string& target,
      const std::string& replacement );
   // replace every instance of text in a string with a different text
   // INPUTS
   // text - text to replace in
   // target - text to be replaced
   // replacement - text substituted for target text
   // RETURN - number of replacements
   // NOTES 
   // 1) Can replace character by itself
   // replace every instance of target in text with replacement
   // 2) Matching is case-sensitive
 
   std::vector<std::string> tokenize( const std::string& text,
      const std::string& delimiters, bool allow_escaped_delimiters=false );
   // break text into tokens (pieces of text separated by delimiters)
   // INPUTS
   // text - text to tokenize
   // delimiters - characters that separate the tokens
   // allow_escaped_delimiters - see below
   // RETURN - tokens (without any delimiters)
   // NOTES 
   // 1) If allow_escaped_delimiters is true, wherever there are two of the
   // same delimiter in a row the code replaces the two by one of those 
   // delimiters and treats that character as text, i.e., adds it to the token
   // If allow_escaped_delimiters is false, it treats the two delimiters as
   // delimiters, not text

   std::string to_upper( std::string text );
   // return capitalized string
   // INPUTS
   // text - string to capitalize
   // RETURN - input with all characters replaced by their capitals,
   // as defined by std::toupper()

   std::string trim( const std::string& text );
   // remove all leading and trailing whitespace
   // INPUTS
   // text - string to trim
   // RETURN - input with all leading and trailing whitespace removed.
   // Whitespace as defined by std::isspace()

   std::vector<std::string> wildcard_file_match( std::string filename=std::string(), 
      std::string path=std::string(), bool case_sensitive = false );
   // Get file names that match a wildcard specification
   // INPUTS
   // filename - the part of the file specification after the directory
   //   e.g., sp97.jo in c:\greg\sp97.jo    
   //   You can use wildcards anywhere in the file name
   // path - No wildcards, can be relative or absolute, can have or omit final directory separator
   //   Forward slash (/) works for both Mac and Windows, back slash (\) works for Windows only
   //   Leave empty or use period to search in current directory
   // case_sensitive - if true match is case sensitive, else case insensitive
   // RETURN - vector of matching file names, each appended to path
   // NOTES 
   // 1) This routine accepts wildcards, not regular expressions. 
   // The two wildcards are the asterisk (*), which matches zero or more characters and
   // the question mark (?), which matches exactly one character
   // 2) calling the function with no arguments does a case-insensitive match to all files in the
   //    current directory

   std::vector<std::string> wildcard_file_match( std::string path_and_file,
      bool case_sensitive = false );
   // Get file names that match a wildcard specification
   // INPUTS
   // path_and_file - a wildcard file specification that may contain any or all of
   // a drive, path, stem and extension, i.e., a normal wildcard specification
   // The part of the file specification after the directory is the file name
   //   e.g., sp97.jo in c:\greg\sp97.jo    
   //   You can use wildcards anywhere in the file name but not in the path (the rest of 
   //   the specification). The path can be relative or absolute, can have or omit final directory separator
   //   Forward slash (/) works for both Mac and Windows, back slash (\) works for Windows only
   //   Leave empty or use period to search in current directory
   // case_sensitive - if true match is case sensitive, else case insensitive
   // RETURN - vector of matching file names, each appended to path
   // NOTES 
   // 1) This routine accepts wildcards, not regular expressions. 
   // The two wildcards are the asterisk (*), which matches zero or more characters and
   // the question mark (?), which matches exactly one character

} // end of namespace FJ

#endif
