#include "run_mnn.hpp"
#include "csv.hpp"

#include "MnnCorrect.hpp"

#include <filesystem>
#include <omp.h>
#include <unordered_map>
#include <unordered_set>

using namespace std;
using namespace FJ;
using namespace std::filesystem;

void prepare_mnn_input( const std::vector< std::vector<std::string> >& stringCsv,
   std::vector<double>& events,std::vector<int>& eventIDs,std::vector<int>& batchIDs );

void save_mnn_output( const MnnSettings& settings,std::vector<double>& correctedParameters,
   int numEvents,int numParameters,std::vector<int>& eventIDs );
   
Summary FJ::run_mnn( const MnnSettings& settings )
{
   string eventsFile = settings.inputFile(0).name;
   std::vector< std::vector<std::string> > stringCsv;
   if( !read_csv( eventsFile, stringCsv, true ) )
   {
      string msg { "Couldn\'t open or read " };
      msg += eventsFile;
      throw runtime_error( msg );
   }
   const int numEvents { static_cast<int>(stringCsv.size()) };
   const int numParameters { static_cast<int>(stringCsv[0].size()-2) }; // last two columns are batch and event IDs
   std::vector<int> batchIDs;
   std::vector<int> eventIDs;
   std::vector<double> parameters;
   prepare_mnn_input( stringCsv, parameters, eventIDs, batchIDs );
      
   mnncorrect::MnnCorrect<> runner;
   runner.set_approximate( settings.approximate() );
   runner.set_automatic_order( settings.automaticOrder() );
   runner.set_mass_cap( settings.massCap() );
   runner.set_num_mads( settings.numMads() );
   runner.set_num_neighbors( settings.numNeighbors() ); // value checked in front end to be [2,200]
   runner.set_robust_iterations( settings.numRobustInterations() );
   runner.set_reference_policy( settings.referencePolicy() );
   runner.set_robust_trim( settings.robustTrim() );

   // Josh wants the number of threads to be half the max to avoid problems
   // they've had in other plugins when using the max number
   int numThreads = omp_get_max_threads() / 2;
   runner.set_num_threads( numThreads );

   std::vector<double> output( numEvents * numParameters );
   runner.run( numParameters, numEvents, parameters.data(), batchIDs.data(), output.data());

   save_mnn_output( settings, output, numEvents, numParameters, eventIDs );

   return Summary(true, numEvents);
}

void prepare_mnn_input( const std::vector< std::vector<std::string> >& stringCsv,
   std::vector<double>& events, std::vector<int>& eventIDs, std::vector<int>& batchIDs )
{
   const int numEvents { static_cast<int>(stringCsv.size()) };
   const int numParameters { static_cast<int>(stringCsv[0].size())-2 };

   // the penultimate column has the batch IDs. Read, convert and store them
   const int batchIx { numParameters };
   batchIDs.resize( numEvents );
   for( int i = 0; i < numEvents; ++i )
      batchIDs[i] = stoi( stringCsv[i][batchIx] );

   // the batch IDs must go from 0 to numParameters-1. Convert them
   unordered_set<int> uniqueIDs;
   for( const auto ID : batchIDs )
      uniqueIDs.insert( ID );

   // if there's only one batch treat that as an error and bail out
   if( uniqueIDs.size() == 1 )
      throw runtime_error( "There must be at least two batches" );

   // make a new batch ID for each original batch ID
   unordered_map<int,int> conversion;
   int ix { 0 };
   for( const auto oldID : uniqueIDs )
      conversion[oldID] = ix++;

   // convert all the old batch IDs
   transform( batchIDs.begin(),batchIDs.end(),batchIDs.begin(),
      [&conversion]( int oldID ) {return conversion[oldID]; } );

   // the last column has the event IDs. Read, convert and store them
   const int eventIx { batchIx+1 };
   eventIDs.resize( numEvents );
   for( int i = 0; i < numEvents; ++i )
      eventIDs[i] = stoi( stringCsv[i][eventIx] );

   // CAUTION - in the input (stringCsv) from FlowJo, each row is an event and
   //   each column is a parameter. The data must be stored in a 1D
   //   array (vector) in column-major order for MnnCorrect. Thus if there are D parameters,
   //   the first D elements of the vector are the parameters for the first event,
   //   the second D elements are the parameters of the second event, etc.
   events.resize( numEvents*numParameters );
   ix = 0;
   for( int i = 0; i < numEvents; ++i )
      for( int j = 0; j < numParameters; ++j )
         events[ix++] = stod( stringCsv[i][j] );
}

void save_mnn_output( const MnnSettings& settings, std::vector<double>& correctedParameters, 
   int numEvents, int numParameters, std::vector<int>& eventIDs )
{
   vector<vector<double>> parameters( numEvents );
   double* inPtr { correctedParameters.data() };
   const int numOutCols { numParameters };
   //const int numOutCols { numParameters+1 }; // +1 because adding event ID at end
   for( int i = 0; i < numEvents; ++i, inPtr+=numParameters)
   {
      parameters[i].reserve( numOutCols );
      parameters[i].assign( inPtr,inPtr+numParameters );
      //parameters[i].push_back( eventIDs[i] );
   }

   // make header. Column titles have the form MNN_n_RUNID
   //    where n     is  1,2,3,...numParameters         and  RUNID is given in the settings
   vector<string> header( numOutCols );
   const string prefix { "MNN_" };
   const string suffix { "_"+settings.runID() };
   for( int i = 0; i < numParameters; ++i )
      header[i] = prefix + to_string( i+1 ) + suffix; // want one-based parameter numbering

   const string eventIDTitle { "EventNumberDP" };
   //header.back() = eventIDTitle;

   if( !write_csv<double>( settings.outputFile(0).name, parameters, header ) )
      throw runtime_error( "Unable to save output to "+settings.outputFile(0).name);
}
